/*
* Copyright (c) 2017 SecureRF Corporation, All rights reserved.
* -------------------------------------------------------------------------
* This computer program is proprietary information of SecureRF Corporation.
* The receipt or possession of this program does not convey any rights to
* reproduce or disclose its contents, or to manufacture, use, or sell
* anything that it may describe, in whole or in part, except for the purposes of
* the post-quantum algorithm public review and evaluation in connection with
* the National Institute of Standards and Technology’s development of standards
* for post-quantum cryptography.  Any reproduction of this program for any other
* purpose without the express written consent of SecureRF is a violation of the
* copyright laws and may subject you to criminal prosecution.
* -------------------------------------------------------------------------
*/


//**************************************************************************
// HEADER
//**************************************************************************

 /**************************************************************************
 * NAME :       ae_crypto.h
 *
 * DESCRIPTION: Algorithms corresponding to those outlined in the
 *              WalnutDSA spec
 *
 */


//**************************************************************************
// INCLUDES
//**************************************************************************

#include <stdint.h>
#include <arpa/inet.h>

/* Library of basic functions */
#include "ae_lib.h"

//**************************************************************************
// DEFINITIONS
//**************************************************************************

/* Number of braid strands. Defined as B10. */
#define WALNUT_BRAID                            10

#if (WALNUT_SECURITY_LEVEL == 128)

/* Size of Galois Field. Defined as M31 (2^31-1) */
#       define WALNUT_FIELD                     0x7FFFFFFF

/* Number of bits of an element in GF(M31) */
#       define GF_ELEM_BITS                     31

/* Size of hashed message for WalnutDSA using SHA256 */
#       define WALNUT_HASH_SIZE                 32

/* Minimal length of certain random braid words
   Defined in section 11.1 */
#       define WALNUT_VALUE_L                   13

/* Minimal length of the private key
   Defined in section 11.1 */
#       define WALNUT_VALUE_l                   124

#endif

#if (WALNUT_SECURITY_LEVEL == 256)

/* Size of Galois Field. Defined as M61 (2^61-1). */
#       define WALNUT_FIELD                     0x1FFFFFFFFFFFFFFF

/* Number of bits of an element in GF(M61) */
#       define GF_ELEM_BITS                     61

/* Size of hashed message for WalnutDSA using SHA512 */
#       define WALNUT_HASH_SIZE                 64

/* Minimal length of certain random braid words
   Defined in section 11.2 */
#       define WALNUT_VALUE_L                   26

/* Minimal length of the private key
   Defined in section 11.2 */
#       define WALNUT_VALUE_l                   275

#endif

/* WALNUT_FIELD being undefined means that WALNUT_SECURITY_LEVEL is invalid */
#if !defined(WALNUT_FIELD)

#       error "INVALID SECURITY LEVEL. VALID OPTIONS: 128, 256"

#endif

/* With Galois Field data, the type chosen must have at least this number of bits */
#define GF_T_BITS                 (GF_ELEM_BITS + (8 - (GF_ELEM_BITS % 8)))

/* With Galois Field data, the type chosen must have at least this number of bytes */
#define GF_T_BYTES                (GF_T_BITS / 8)

/* Determine Galois Field data type */
#if   GF_T_BYTES == 1
#       define  GF_T         uint8_t
#       define  htongft(x)   (x)
#       define  ntohgft(x)   (x)
#       define GF_STRFMT    "x"
#elif GF_T_BYTES == 2
#       define  GF_T         uint16_t
#       define  htongft(x)   (htons(x))
#       define  ntohgft(x)   (ntohs(x))
#       define GF_STRFMT    "x"
#elif GF_T_BYTES <= 4
#       define  GF_T         uint32_t
#       define  htongft(x)   (htonl(x))
#       define  ntohgft(x)   (ntohl(x))
#       define GF_STRFMT    "08x"
#elif GF_T_BYTES <= 8
#       define  GF_T         uint64_t
#       define  htongft(x)   (htonll(x))
#       define  ntohgft(x)   (ntohll(x))
#       define GF_STRFMT    "llx"
#else
#       error   "INVALID GF_ELEM_BITS VALUE"
#endif

/* Number of bits to define a strand in B10 */
#define B10_STRAND_BITS                         0x04

/* Number of bits to define a generator in B10 */
#define B10_GENERATOR_BITS                      (B10_STRAND_BITS + 1)

/* Maximum length of a pure braid */
#define B10_MAX_PURE_BRAID_LENGTH               0x14

/* Maximum length of a cloaking element (2K) */
#define B10_MAX_CLOAKING_ELEMENT_LENGTH         0x800

/* Maximum length of any braid (16K) */
#define B10_MAX_BRAID_LENGTH                    0x4000

/* Number of T-values within the public key */
#define WALNUT_NUM_TVALUES                      WALNUT_BRAID

/* Number of matrix elements within the public key */
#define WALNUT_NUM_MATRIX_ELEMENTS              (WALNUT_BRAID * WALNUT_BRAID)

/* Number of permutation elements within the public key */
#define WALNUT_NUM_PERMUTATION_ELEMENTS         WALNUT_BRAID

/* Number of bits per T-value */
#define WALNUT_TVALUE_BITS                      GF_T_BITS

/* Number of bits per matrix element */
#define WALNUT_MATRIX_ELEMENT_BITS              GF_T_BITS

/* Number of bits per permutation element */
#define WALNUT_PERMUTATION_ELEMENT_BITS         B10_STRAND_BITS

/* Public key version information in three bytes:
   one byte for number of braid strands
   two bits to determine whether or not this is a Mersenne prime
   remaining bits to determine field size */
#define WALNUT_PUBKEY_INFORMATION               ((WALNUT_BRAID << 16) | (((WALNUT_FIELD >> 14) & 1) << 14) | GF_ELEM_BITS)

/* Number of bits within the public key used to track version/key information (padded to fill a GF_T element) */
#define WALNUT_VERSION_BITLENGTH                ((0x03 * 8) + (GF_T_BITS - ((0x03 * 8) % GF_T_BITS)))

/* Number of bits within the public key that contain T-Values */
#define WALNUT_PUBKEY_TVALUES_BITLENGTH         (WALNUT_NUM_TVALUES * WALNUT_TVALUE_BITS)

/* Number of bits within the public key that contain matrix elements */
#define WALNUT_PUBKEY_MATRIX_BITLENGTH          ((WALNUT_NUM_MATRIX_ELEMENTS - WALNUT_BRAID + 1) * WALNUT_MATRIX_ELEMENT_BITS)

/* Number of bits within the public key that contain matrix elements */
#define WALNUT_PUBKEY_PERMUTATION_BITLENGTH     (WALNUT_NUM_PERMUTATION_ELEMENTS * WALNUT_PERMUTATION_ELEMENT_BITS)

/* Number of bytes within the public key used to track version/key information */
#define WALNUT_VERSION_LENGTH                   ((WALNUT_VERSION_BITLENGTH + 7) / 8)

/* Maximum number of bytes within the public key that contain T-Values */
#define WALNUT_PUBKEY_TVALUES_LENGTH            ((WALNUT_PUBKEY_TVALUES_BITLENGTH + 7) / 8)

/* Maximum number of bytes within the public key that contain matrix elements (+7 used so that value is rounded up on division) */
#define WALNUT_PUBKEY_MATRIX_LENGTH             ((WALNUT_PUBKEY_MATRIX_BITLENGTH + 7) / 8)

/* Number of bytes within the public key per row of matrix elements */
#define WALNUT_PUBKEY_MATRIX_ROW_LENGTH         ((WALNUT_BRAID * WALNUT_MATRIX_ELEMENT_BITS + 7) / 8)

/* Number of bytes within the public key per column of matrix elements */
#define WALNUT_PUBKEY_MATRIX_COL_LENGTH         WALNUT_PUBKEY_MATRIX_ROW_LENGTH

/* Maximum number of bytes within the public key that contain matrix elements */
#define WALNUT_PUBKEY_PERMUTATION_LENGTH      (((WALNUT_NUM_PERMUTATION_ELEMENTS * WALNUT_PERMUTATION_ELEMENT_BITS) + 7) / 8)

/* Number of elements that constitute the public key */
#define WALNUT_PUBKEY_ELEMENTS                  ((WALNUT_VERSION_LENGTH + \
                                                  WALNUT_PUBKEY_TVALUES_LENGTH + \
                                                  WALNUT_PUBKEY_MATRIX_LENGTH + \
                                                  WALNUT_PUBKEY_MATRIX_LENGTH + \
                                                  WALNUT_PUBKEY_PERMUTATION_LENGTH + \
                                                  GF_T_BYTES - 1) / GF_T_BYTES)

/* Number of bytes that constitute the public key */
#define WALNUT_PUBKEY_LENGTH                     (WALNUT_PUBKEY_ELEMENTS * GF_T_BYTES)

/* Position of the version within both the public key and the signature */
#define WALNUT_VERSION_POSITION                 0x00

/* Position of the T-values within the public key */
#define WALNUT_PUBKEY_TVALUES_POSITION          (WALNUT_VERSION_POSITION + (WALNUT_VERSION_LENGTH / GF_T_BYTES))

/* Position of the S matrix elements within the public key */
#define WALNUT_PUBKEY_S_MATRIX_POSITION         (WALNUT_PUBKEY_TVALUES_POSITION + \
                                                  (WALNUT_PUBKEY_TVALUES_LENGTH / GF_T_BYTES))

/* Position of the S' matrix elements within the public key */
#define WALNUT_PUBKEY_S_PRIME_MATRIX_POSITION   (WALNUT_PUBKEY_S_MATRIX_POSITION + \
                                                  (WALNUT_PUBKEY_MATRIX_LENGTH / GF_T_BYTES))

/* Position of the S permutation elements within the public key (at the end of the public key) */
#define WALNUT_PUBKEY_S_PERMUTATION_POSITION    (WALNUT_PUBKEY_S_PRIME_MATRIX_POSITION + \
                                                  (WALNUT_PUBKEY_MATRIX_LENGTH / GF_T_BYTES))

//**************************************************************************
// MACROS
//**************************************************************************

/* Given braid and position, get generator at that position.
   Takes into account offset of 2, as the first two braid elements are its length */
#define GET_BRAID_GENERATOR(braid, pos)         (braid[pos + 2])

/* Given bands and position, get band a_(t,s) at that position */
#define GET_BAND_GENERATOR(bands, Ai, pos)      (bands[(Ai)+1][(pos)])

/* Given bands and position, get band a_(t,s) at that position, then extract t */
#define GET_BAND_GENERATOR_T(bands, Ai, pos)    ((((GET_BAND_GENERATOR(bands, Ai, pos)) & 0xFF00)) >> 8)

/* Given bands and position, get band a_(t,s) at that position, then extract s */
#define GET_BAND_GENERATOR_S(bands, Ai, pos)    (((GET_BAND_GENERATOR(bands, Ai, pos)) & 0x00FF))

/* Given generator gen, get its strand (absolute value of generator) */
#define GET_GENERATOR_STRAND(gen)               ((gen) & 0x0F)

/* Given generator gen, get its sign */
#define GET_GENERATOR_SIGN(gen)                 ((gen) & 0x10)

/* Given braid, get its number of generators */
#define GET_NUM_BRAID_GENERATORS(braid)         ((braid[0] << 8) | braid[1])

/* Given bands, get its number of band generators */
#define GET_NUM_BAND_GENERATORS(bands)          ((bands[0][0] << 8) | bands[0][1])

/* Given braid, get its number of bytes */
#define GET_NUM_BRAID_BYTES(braid)              ((GET_NUM_BRAID_GENERATORS(braid)) + 2)

/* Given braid, get its number of bytes when packed */
#define GET_NUM_BRAID_BYTES_PACKED(braid)       ((((GET_NUM_BRAID_GENERATORS(braid) * B10_GENERATOR_BITS) + 7) / 8) + 2)

/* Given braid, generator, and position, set generator at position pos in braid to gen (braid[pos] = generator) */
#define SET_BRAID_GENERATOR(braid, gen, pos)    ((braid[(pos) + 2]) = (gen))

/* Given bands, generator, and position, set generator at position pos in bands */
#define SET_BAND_GENERATOR(bands, gen, Ai, pos) (bands[(Ai)+1][(pos)] = gen)

/* Given t, s, and sign, make a band generator of form a_(t,s)^sign (where sign = 0, 1) */
#define MAKE_BAND_GENERATOR(t, s, sign)         (((((t) & 0x0F) | ((!!sign) << B10_STRAND_BITS)) << 8) \
                                                  | (((s) & 0x0F) | ((!!sign) << B10_STRAND_BITS)))

/* Given a GF(Mx) two-argument function, and two operands a and b, find a op b in GF(Mx) */
#define GOP_BASE(function, a, b)               (((a > WALNUT_FIELD-1) || (b > WALNUT_FIELD-1)) ? 0 : function(a, b))

/* Given a GF(Mx) one-argument function, and one operand a, find op(a) in GF(Mx) */
#define MOP_BASE(function, a)                  ((a > WALNUT_FIELD-1) ? 0 : function(a))

/* If WALNUT_SECURITY_LEVEL is 128, then use lookup tables in GF(M31) */
#if WALNUT_SECURITY_LEVEL == 128

#       define GADD(a, b)       (GOP_BASE(gadd_m31, (a), (b)))
#       define GMUL(a, b)       (GOP_BASE(gmul_m31, (a), (b)))
#       define MINV(a)          (MOP_BASE(minv_m31, (a)))
#       define MNEG(a)          (MOP_BASE(mneg_m31, (a)))

#endif

/* If WALNUT_SECURITY_LEVEL is 256, then use lookup tables in GF(M61) */
#if WALNUT_SECURITY_LEVEL == 256

#       define GADD(a, b)       (GOP_BASE(gadd_m61, (a), (b)))
#       define GMUL(a, b)       (GOP_BASE(gmul_m61, (a), (b)))
#       define MINV(a)          (MOP_BASE(minv_m61, (a)))
#       define MNEG(a)          (MOP_BASE(mneg_m61, (a)))

#endif

//**************************************************************************
// FUNCTIONS
//**************************************************************************

/**************************************************************************
* NAME :       int key_generation(*privkey, *pubkey)
*
* DESCRIPTION: Generate both the public and the private key
*
* ARGUMENTS:
*
*      INPUTS:
*              void            *privkey        buffer to store
*                                              private key
*              void            *pubkey         buffer to store
*                                              public key
*
*      OUTPUTS:
*              void            *privkey        buffer containing
*                                              private key
*              void            *pubkey         buffer containing
*                                              public key
*              int             return          on success, return 1
*
*/
int key_generation(void *privkey, void *pubkey);

/**************************************************************************
* NAME :       int signature_generation(signature, sig_length, message, length, privkey)
*
* DESCRIPTION: Generate the signature
*
* ARGUMENTS:
*
*      INPUTS:
*              void            *signature      buffer for signature and message
*              void            *sig_length     pointer to signature length
*              const void      *message        message
*              size_t          length          length of message
*              const void      *privkey        private key
*
*      OUTPUTS:
*              void            *signature      buffer containing signature and message
*              void            *sig_length     pointer to signature length
*              int             return          on success, return 1
*
*/
int signature_generation(void *signature, void *sig_length, const void *message, size_t length, const void *privkey);

/**************************************************************************
* NAME :       int signature_verification(pubkey, signature, sig_length)
*
* DESCRIPTION: Verify the signature
*
* ARGUMENTS:
*
*      INPUTS:
*              void            *pubkey         public key
*              const void      *signature      buffer containing signature and message
*              void            *sig_length     pointer to length of signature buffer
*              unsigned char   **returnmes     pointer for message returned to SUPERCOP
*              unsigned long long     *meslen  length of the message
*
*      OUTPUTS:
*              unsigned char   **returnmes     pointer for message returned to SUPERCOP
*              int             return          on success, return 1
*
*/
int signature_verification(void *pubkey, const void *signature, void *sig_length, unsigned char **returnmes, unsigned long long *meslen);

#ifdef KAT_INTERMEDIATES
void unpack_and_print_braid(uint8_t *braid, unsigned char *name);
void printBuffer(GF_T *braid, size_t size,  char *name);
void printCharBuffer(uint8_t *braid, size_t size,  char *name);
void unpack_and_print_matrix(GF_T *matrix,  char *name);
#endif
