/*
* Copyright (c) 2017 SecureRF Corporation, All rights reserved.
* -------------------------------------------------------------------------
* This computer program is proprietary information of SecureRF Corporation.
* The receipt or possession of this program does not convey any rights to
* reproduce or disclose its contents, or to manufacture, use, or sell
* anything that it may describe, in whole or in part, except for the purposes of
* the post-quantum algorithm public review and evaluation in connection with
* the National Institute of Standards and Technology’s development of standards
* for post-quantum cryptography.  Any reproduction of this program for any other
* purpose without the express written consent of SecureRF is a violation of the
* copyright laws and may subject you to criminal prosecution.
* -------------------------------------------------------------------------
*/


#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include "ae_lib.h"

#ifdef WORKING_COPY
#include "rng.h"
#else
#include "randombytes.h"
#endif

//**************************************************************************
// FUNCTIONS
//**************************************************************************
uint32_t gadd_m31(uint32_t a, uint32_t b)
{
    uint32_t ret;

    /* Unreduced sum of a and b */
    ret = a + b;

    /* Reduce sum to a value in GF(M31) by computing (a + b) mod M31 */
    ret = ret % 0x7FFFFFFF;

    return ret;
}

uint32_t gmul_m31(uint32_t a, uint32_t b)
{
    uint64_t ret;

    /* Unreduced product of a and b */
    ret = (uint64_t)a * (uint64_t)b;

    /* Reduce sum to a value in GF(M31) by computing (a * b) mod M31 */
    ret = ret % 0x7FFFFFFF;

    return (uint32_t)ret;
}

uint32_t minv_m31(uint32_t arg)
{
    uint32_t a, b, y, z, i;

    a = 1;
    b = 0;
    y = arg;
    z = 0x7FFFFFFF;

    if(arg == 0) {
        return 0;
    }

    for(;;) {

        i = 31;
        while(!(y & 1)) {
            y >>= 1;
            i--;
        }

        a = ((a << i) | (a >> (31-i))) & 0x7FFFFFFF;

        if(y == 1) {
            return a;
        }


        i = a;
        a = gadd_m31(a, b);
        b = i;

        i = y;
        y = y + z;
        z = i;
    }
}

uint32_t mneg_m31(uint32_t a)
{
    /* For a,-a ϵ GF(prime), -a is defined by (a + -a) mod p = 0 */
    /* In a prime field p, a + -a = p because p mod p = 0 */
    return 0x7FFFFFFF - a;
}

uint64_t gadd_m61(uint64_t a, uint64_t b)
{
    uint64_t ret;

    /* Unreduced sum of a and b */
    ret = a + b;

    /* Reduce sum to a value in GF(M31) by computing (a + b) mod M31 */
    ret = ret % 0x1FFFFFFFFFFFFFFF;

    return ret;
}

uint64_t gmul_m61(uint64_t a, uint64_t b)
{
    /* Because there isn't a portable uint128_t data type, we will instead use the
       FOIL method to calculate a * b with 32-bit operands, which will provide 64-bit products */

    /* hi/lo: upper/lower 61 bits of product */
    uint64_t hi, lo;

    /* if a = cx + j and b = dx + k, then
       a * b = (cx + j)(dx + k) = (cd)x^2 + (ck + jd)x + (jk)
       below: F = cd, O = ck, I = jd, L = jk */
    uint64_t f, o, i, l;

    /* We want the "cutoff" between the lo/hi 61 bits to be at a 64 bit boundary
       Shifting left three will do this, but note the lo 61 bit result will be offset left by 3 */
    a <<= 3;

    /* Calculate the "F" in "FOIL" */
    f = (a >> 32) * (b >> 32);
    /* Calculate the "O" in "FOIL" */
    o = (a >> 32) * (b & 0xFFFFFFFF);
    /* Calculate the "I" in "FOIL" */
    i = (a & 0xFFFFFFFF) * (b >> 32);
    /* Calculate the "L" in "FOIL" */
    l = (a & 0xFFFFFFFF) * (b & 0xFFFFFFFF);

    /* Calculate the sum as follows: */
    /* f = "F"OIL, o = F"O"IL, i = FO"I"L, l = FOI"L"
     *
     *          ffffffffffffffff0000000000000000
     *        + 00000000oooooooooooooooo00000000
     *        + 00000000iiiiiiiiiiiiiiii00000000
     *        + 0000000000000000llllllllllllllll
     *        = |------hi------||------lo------|
     */
    lo = l + (o << 32);
    hi = (lo < (o << 32));

    lo += (i << 32);
    hi += (lo < (i << 32));

    hi += f + (o >> 32) + (i >> 32);

    // fix the left 3 offset by shifting right 3
    lo >>= 3;

    return gadd_m61(hi, lo);
}

uint64_t minv_m61(uint64_t arg)
{
    uint64_t a, b, y, z, i;

    a = 1;
    b = 0;
    y = arg;
    z = 0x1FFFFFFFFFFFFFFF;

    if(arg == 0) {
        return 0;
    }

    for(;;) {

        i = 61;
        while(!(y & 1)) {
            y >>= 1;
            i--;
        }

        a = ((a << i) | (a >> (61-i))) & 0x1FFFFFFFFFFFFFFF;

        if(y == 1) {
            return a;
        }


        i = a;
        a = gadd_m61(a, b);
        b = i;

        i = y;
        y = y + z;
        z = i;
    }
}

uint64_t mneg_m61(uint64_t a)
{
    /* For a,-a ϵ GF(p), -a is defined by (a + -a) mod p = 0 */
    /* In a prime field p, a + -a = p because p mod p = 0 */
    return 0x1FFFFFFFFFFFFFFF - a;
}

uint8_t get_least_common_multiple(uint8_t *elem, size_t length)
{
    /* Loop counters */
    size_t i, j;

    /* If there are no elements, return 0 */
    if(length == 0) {
        return 0;
    }

    /* Initialize greatest common denominator to 1 */
    uint8_t gcd = 1;

    /* Initialize answer to first element in elem */
    uint8_t ans = elem[0];

    for(i=1; i<length; i++) {

        /* Get greatest common denominator */
        for(j=1; (j<=elem[i] && j<=ans); j++) {
            if((elem[i] % j == 0) && (ans % j == 0)) {
                gcd = j;
            }
        }

        /* Find least common multiple */
        ans = (elem[i] * ans) / gcd;

        /* Reinitialize greatest common denominator to 1 */
        gcd = 1;
    }

    return ans;
}

#if WALNUT_SECURITY_LEVEL == 128
uint32_t get_strong_random_number(uint32_t max)
{
    /* Initialize the random number to zero */
    uint64_t number = 0;

    /* Extract four random bytes from randombytes */
    randombytes((uint8_t *)(&number), 8);

    /* Return number mod max (if max is zero, then just return the number) */
    return (max != 0) ? (number % max) : number;
}
#elif WALNUT_SECURITY_LEVEL == 256
uint64_t get_strong_random_number(uint64_t max)
{
    /* Initialize the random number to zero */
    uint64_t number = 0;

    /* Extract four random bytes from randombytes */
    randombytes((uint8_t *)(&number), 8);

    /* Return number mod max (if max is zero, then just return the number) */
    return (max != 0) ? (number % max) : number;
}
#endif

uint8_t get_permutation_order(uint8_t *p, size_t num)
{
    /* Loop counters */
    size_t i, j;

    /* Product of disjoint cycles */
    uint8_t pdc[num];

    /* Cycle lengths */
    uint8_t cycle_lengths[num];

    /* Initialize all cycle_lengths to zero */
    memset(cycle_lengths, 0, num);

    /* Get product of disjoint cycles */
    get_product_of_disjoint_cycles(pdc, p, num);

    /* Build the array of cycle lengths */
    for(i=0, j=0; i<num; i++) {

        /* Increment the cycle_length of cycle k */
        cycle_lengths[j]++;

        /* If this is the end of a cycle, then move to the next cycle in cycle_length */
        if(pdc[i] & 0x80) {
            j++;
        }
    }

    /* Return least common multiple of the permutation cycle lengths, which is equivalent to the order */
    return get_least_common_multiple(cycle_lengths, j);
}

uint8_t generate_random_permutation(uint8_t *p, size_t num)
{
    /* Loop counter */
    uint8_t i;

    /* 8.1: Start with the identity permutation of n elements */
    for(i=0; i<num; i++) {
        p[i] = i;
    }

    /* Temporary value holder */
    uint8_t tmp;

    /* Random location */
    uint8_t j;

    /* Shuffle the identity permutation */
    /* 8.2: Start with the last (1-indexed) offset, i = n
            For zero-indexed, i = n-1 */
    for(i=num-1; i!=0; i--) {

        /* 8.3: Choose a random number 1 ≤ j ≤ i */
        j = get_strong_random_number(i+1);

        /* 8.4: Swap permutation elements i and j */
        tmp = p[i];
        p[i] = p[j];
        p[j] = tmp;

        /* 8.5: Iterate i down to 1 */
    }

    return 1;
}

uint8_t generate_random_permutation_of_given_order(uint8_t *p, size_t num, uint8_t order)
{
    /* Track success */
    uint8_t success = 1;

    /* Loop counters */
    size_t i, j;

    /* Maximum attempts */
    size_t max_attempts = 5000;

    /* If the order input is zero, then fail. */
    if(order == 0) {
        return 0;
    }

    /* While the permutation does not have the given order (up to max_attempts attempts) */
    for(i=0, j=0; i<max_attempts; i++) {

        /* Generate a random permutation */
        success &= generate_random_permutation(p, num);

        /* Get the permutation order */
        j = get_permutation_order(p, num);

        /* If j == order, then a permutation with the given order has been found. */
        if(j == order) {

            /* Done */
            return success;
        }
    }

    /* If the loop return wasn't reached, then a permutation of order wasn't found. */
    return 0;
}

uint8_t invert_permutation(uint8_t *p, size_t num)
{
    /* Loop counter */
    size_t i;

    /* Inverse permutation */
    uint8_t inv[num];

    /* Invert the permutation */
    for(i=0; i<num; i++) {
        inv[p[i]] = i;
    }

    /* Overwrite the original permutation with its inverse */
    memcpy(p, inv, num);

    return 1;
}

uint8_t multiply_permutations(uint8_t *p3, uint8_t *p1, uint8_t *p2, size_t num)
{
    /* Loop counter */
    size_t i;

    /* Multiply p1 by p2 */
    for(i=0; i<num; i++) {
        p3[i] = p1[p2[i]];
    }

    return 1;
}

uint8_t compare_identitiy_permutation(uint8_t *p1, size_t num)
{
    /* Loop counter */
    size_t i;

    /* Identity permutation */
    uint8_t identity[num];

    /* Generate identity permutation */
    for(i=0; i<num; i++) {
        identity[i] = i;
    }

    /* Compare the two permutations */
    /* Return 1 on equality; return 0 on inequality */
    return !memcmp(p1, identity, num);
}

uint8_t get_product_of_disjoint_cycles(uint8_t *pdc, uint8_t *permutation, uint8_t length)
{
    /* Loop counters */
    uint8_t i, j;

    /* Initialize the product of disjoint cycles as the identity permutation */
    for(i=0; i<length; i++) {
        pdc[i] = i;
    }

    /* First element to be checked */
    uint8_t starting_element = pdc[0];

    /* Cycle element */
    uint8_t cycle_element = pdc[0];

    /* 9.1(a): First write σ as a product of disjoint cycles C1 ... Cn */
    for(i=0; i<length; i++) {

        /* If the starting element is equal to neither the braid permutation at the starting element
           nor the braid permutation at the cycle element, then this is within a cycle */
        if(permutation[starting_element] != starting_element && permutation[cycle_element] != starting_element) {

            /* Following code adjusts the product of disjoint cycles pdc to track used elements
               For example, given pdc initialized to the identity permutation: (0 1 2 3 4 5 6 7)
               If i == 0 and permutation[i] == 5, pdc is changed to:           (0 5 1 2 3 4 6 7) */

            /* Store the cycle element */
            cycle_element = permutation[pdc[i]];

            /* Find the location of the cycle element within the product of disjoint cycles */
            for(j=i; j<length; j++) {
                if(pdc[j] == cycle_element) {
                    break;
                }
            }

            /* Shift elements right by one */
            memmove(&pdc[i+1], &pdc[i], j-i);

            /* Place cycle element in its new position */
            pdc[i+1] = cycle_element;

        /* Else, this element either maps to itself or is the end of a cycle */
        } else {

            /* Mark this element */
            pdc[i] |= 0x80;

            /* Set the starting element for the next iteration (check that i+1 is not out of bounds) */
            starting_element = (i+1 < length) ? pdc[i+1] : 0;
        }
    }

    /* Adjust the cycle to match form listed in the WalnutDSA specifications.
       Currently, cycles C1...Cs are in the correct order (9.2(b)), but the smallest element is listed first.
       To put it last, step through each cycle and slide the first element to the last position. */
    for(i=0, j=0, starting_element=pdc[0]; i<length; i++) {

        /* If this is the end of a cycle */
        if(pdc[i] & 0x80) {

            /* Clear the end of cycle denotation */
            pdc[i] &= ~0x80;

            /* Shift cycle elements left by one */
            memmove(&pdc[j], &pdc[j+1], i-j);

            /* Set the new end of cycle element to the starting element */
            pdc[i] = (starting_element | 0x80);

            /* Set the starting element for the next iteration (check that i+1 is not out of bounds) */
            starting_element = (i+1 < length) ? pdc[i+1] : 0;

            /* Set the starting element location for the next iteration */
            j = i+1;
        }
    }

    return 1;
}

uint8_t trim_product_of_disjoint_cycles(uint8_t *pdc, uint8_t init_length, uint8_t *final_length)
{
    /* Loop counters */
    uint8_t i, j;

    /* For each cycle element */
    for(i=0, j=0; i<init_length; i++) {

        /* A cycle element must be trimmed if it is a cycle of length 1.
           All other cycle elements must be preserved; for each, one of the two following conditions must be met:
            - MSB of pdc[i] is not set
            - MSB of pdc[i-1] is not set, where i!=0 */
        if((!(pdc[i] & 0x80)) | ((i != 0) && (!(pdc[i-1] & 0x80)))) {

            /* Keep this cycle element */
            pdc[j] = pdc[i];

            /* Increment j */
            j++;
        }
    }

    /* Store the new length */
    *final_length = j;

    return 1;
}

uint8_t sort_product_of_disjoint_cycles(uint8_t *pdc, uint8_t *length)
{
    /* Loop counters */
    size_t i, j;

    /* Starting index of a cycle */
    uint8_t start;

    /* Maximum value encountered within the cycle */
    uint8_t max;

    /* Index of maximum value encountered within the cycle */
    uint8_t max_index;

    /* Temporary array of elements */
    uint8_t tmp[(*length)];

    for(i=0, max=0, max_index=0, start=0; i<(*length); i++) {

        /* If this cycle element is greater than or equal to the recorded max */
        if((pdc[i] & ~0x80) >= max) {

            /* Store this value and its index */
            max = (pdc[i] & ~0x80);
            max_index = i - start;
        }

        /* Copy this cycle element to tmp */
        tmp[i - start] = (pdc[i] & ~0x80);

        /* If this is the end of a cycle */
        if(pdc[i] & 0x80) {

            /* Iterate over all elements of the cycle */
            for(j=0; j<=(i - start); j++) {

                /* Copy elements in correct order.
                   First element copied from tmp will be the greatest number in the cycle. */
                pdc[j + start] = tmp[(j + max_index) % (i - start + 1)];
            }

            /* Mark the last element as the end of the cycle */
            pdc[i] |= 0x80;

            /* Next cycle will start at i+1 */
            start = i+1;

            /* Reset max */
            max = 0;
        }
    }

    return 1;
}

uint8_t reverse_product_of_disjoint_cycles(uint8_t *pdc, uint8_t *length)
{
    /* Loop counters */
    size_t i, j;

    /* Product of disjoint cycles (local to this function) */
    uint8_t pdc_local[*length];

    /* Index where each cycle ends */
    int8_t index[*length+1];

    /* Copy pdc to pdc_local */
    memcpy(pdc_local, pdc, *length);

    /* For every element of pdc_local */
    for(i=0, j=0, index[0]=-1; i<(*length); i++) {

        /* If this element marks the end of a cycle */
        if(pdc_local[i] & 0x80) {

            /* Save this location, and increment to the next cycle */
            index[++j] = i;
        }
    }

    /* For every element of pdc_local */
    for(i=0, j=1; i<(*length); i++) {

        /* In reverse order, copy cycles from pdc_local into pdc */
        pdc[(*length) - i - 1] = pdc_local[index[j] + index[j-1] - i + 1];

        /* If this is the end of a cycle */
        if((index[j] + index[j-1] - i + 1) == (index[j-1] + 1)) {

            /* Move to the next cycle */
            j++;
        }
    }

    return 1;
}

uint8_t generate_integer_partition(uint8_t *partition, uint32_t *num_partition, uint16_t total, uint8_t min, uint8_t max)
{
    /* Randomly generated element */
    uint8_t element;

    /* Because the running total can drop lower than zero, make it signed. */
    int running_total = (int)total;

    /* Index where running_total was last seen greater than min*max */
    uint16_t last_index = 0;

    /* Initialize num_partition to zero */
    *num_partition = 0;

    /* While the running total is not equal to 0 */
    while(running_total != 0) {

        /* 5.3.2.3(a)(1): Generate a random number between min and max (inclusive) */
        element = get_strong_random_number(max - min + 1) + min;

        /* Store this element to the partition */
        partition[*num_partition] = element;

        /* 5.3.2.3(a)(2): Subtract from the initial total. */
        running_total -= element;

        /* 5.3.2.3(b): If the running total is zero, then the partition has been found.
           In this case, neither the following "if" nor the following "else if" statement will evaluate true. */

        /* However, if the running total has dropped below the minimum (cast min to int to avoid sign issues),
           and the running total is NOT equal to zero, then it is not possible to cleanly partition the running
           total such that min<=value<=max for each value. */
        if((running_total < (int)min) && running_total != 0) {

            /* 5.3.2.3(c)(1): Jump back to the position in the list where the running total was greater
               than minimum * maximum. */
            *num_partition = last_index;

            /* 5.3.2.3(c)(2): Set the running total to this value. */
            running_total = (int)total;

        /* Else if the running total is greater than min*max */
        } else if(running_total > (int)(min*max)) {

            /* Mark this index as the most recent index where running_total is greater than min*max */
            last_index = *num_partition;

            /* Subtract this element from the overall total */
            total -= element;
        }

        /* Increment the total number of partitions */
        (*num_partition)++;

        /* 5.3.2.3(d): Repeat until the running total becomes zero. */
    }

    return 1;
}

uint8_t extract_elem(const uint8_t *in, size_t curr_bit, uint8_t packwidth)
{
    /* Element to be returned */
    uint8_t ret = 0;

    /* The mask used to return the extracted element */
    uint8_t mask = 0xFF >> (8 - packwidth);

    /* The amount by which to shift the byte(s) of input buffer in */
    int8_t shiftby = (8 - packwidth) - (curr_bit % 8);

    /* curr_bit is the location of the first bit
       addr is the location of the byte that contains this bit */
    size_t addr = curr_bit / 8;

    /* If shiftby is negative, the bits needed for the extracted element are
       located at both the current buffer address and the next buffer
       address */
    if(shiftby < 0) {

        /* Entry at current buffer address */
        uint8_t entry1 = in[addr];

        /* Entry at incremented buffer address */
        uint8_t entry2 = in[addr + 1];

        /* shiftby for entry2 */
        uint8_t t = 8 + shiftby;

        /* Extracted bits from entry1, shifted into the proper place */
        ret = (entry1 << -shiftby);

        /* Extracted bits from entry2, shifted into the proper place */
        ret |= (entry2 >> t) & mask;

    /* Else, the bits are all located at the current buffer address */
    } else {

        /* Entry at current buffer address */
        uint8_t entry = in[addr];

        /* Shift the entry */
        ret = (entry >> shiftby);
    }

    return ret & mask;
}

uint8_t pack_elem(void *buf, uint8_t packwidth, size_t num_elem)
{
    /* If packwidth is outside the range of 1 and 8 inclusive, then fail */
    if(packwidth < 1 || packwidth > 8) {
        return 0;
    }

    /* Loop counter */
    uint16_t i;

    /* Pointer to buffer */
    uint8_t * buf_ptr = buf;

    /* Element to be packed */
    uint8_t elem;

    /* The amount by which to shift the bit(s) of elem */
    int8_t shiftby = 0;

    /* New storage location for elem */
    size_t addr = 0;

    /* Used to clear bits at the new storage location for elem */
    uint8_t clr_bits = 0xFF >> (8 - packwidth);

    /* Iterate through total number of bits (not bytes) */
    for(i=0; i<(num_elem * packwidth); i+=packwidth) {

        /* Determine the shift amount */
        shiftby = (8 - packwidth) - (i % 8);

        /* Convert bits to bytes, as addresses are in bytes */
        addr = i / 8;

        /* Get element from unpacked location */
        elem = buf_ptr[i / packwidth];

        /* If shiftby is negative, elem must be stored in two pieces
           The upper portion will be stored at buf[addr], and the lower portion
           will be stored at buf[addr+1] */
        if(shiftby < 0) {

            /* Clear the bits at the new storage location */
            buf_ptr[addr] &= ~(clr_bits >> -shiftby);
            buf_ptr[addr+1] &= ~(clr_bits << (shiftby + 8));

            /* Place elem into buf */
            buf_ptr[addr] |= elem >> -shiftby;
            buf_ptr[addr+1] |= elem << (shiftby + 8);

        /* Else, the bits will all be stored at the current buffer address */
        } else {

            /* Clear the bits at the new storage location */
            buf_ptr[addr] &= ~(clr_bits << shiftby);

            /* Place elem into buf */
            buf_ptr[addr] |= elem << shiftby;
        }
    }

    /* Clear data at unused element positions */
    if((shiftby < 0) && (i != 0)) {

        /* Clear unused portion of last modified byte */
        buf_ptr[addr+1] &= ~(0xFF >> -shiftby);

        memset(&buf_ptr[addr+2], 0, num_elem - (addr+2));
    } else {

        /* Clear unused portion of last modified byte */
        buf_ptr[addr] &= ~(0xFF >> (8 - shiftby));

        memset(&buf_ptr[addr+1], 0, num_elem - (addr+1));
    }

    return 1;
}

#ifdef WORKING_COPY
uint8_t sha2_256(void *data, size_t length, uint8_t *hash)
{
    /* Pointer to data */
    uint8_t * d = data;

    SHA256(d, length, hash);

    return 1;
}

uint8_t sha2_512(void *data, size_t length, uint8_t *hash)
{ 
    /* Pointer to data */
    uint8_t * d = data;

    SHA512(d, length, hash);

    return 1;
}
#endif
